import java.util.*;


public class MDP {

	State States[];
	Action PossibleActions[];
	Queue<Action> TakenActions;
	CostAndReward C;
	TransitionMatrix M;
	private double V[];
	private Action Policy[];
	private double Residual[];
	private State CurrentState;
	private State InitialState;
	private int numActionsTaken;
	private final double EPSILON = 0.0000001;
	private final double gamma = 0.2;
	//private double n;
	
	MDP(State S[], State initialState, Action a[], TransitionMatrix matriz, CostAndReward c){
		//Constructor...
		States = S;
		PossibleActions = a;
		C = c;
		M = matriz;
		CurrentState = initialState;
		InitialState = initialState;
		V = new double[States.length];
		Policy = new Action[States.length];
		Residual = new double[States.length];
		numActionsTaken = 0;
		TakenActions = new LinkedList<Action>();
	}
	
	void initializeV(){
		for (int i=0; i < this.States.length; i++){
			if(this.States[i].isMeta){
				// Meta State: must initialize with 0.
				this.V[i] = 0;
			}
			else{
				this.V[i] = 100.0;
				//No exemplo bobo com V[i] = 1, entra em loop!
				
			}
		}	
	}
	
	Action refreshV(){
		double min = Double.MAX_VALUE;
		int minAction = 0;
		
		
		for(int AllStates=0; AllStates < this.States.length; AllStates++){
			
			for(int i=0; i < this.PossibleActions.length; i++){
				double calc = 0;
			
				for(int j=0; j < this.States.length; j++){
					calc += M.Matrix[i][AllStates][j] * this.V[j]; 
				}
				calc = calc * this.gamma; 
				calc += C.Matrix[i][AllStates];
				if(calc < min){
					min = calc;
					minAction = i;
				}
			}
		
			this.Residual[AllStates] = Math.abs(this.V[AllStates] - min);
			//System.out.println(this.Residual[AllStates]);
			this.V[AllStates] = min;
			this.Policy[AllStates] = this.PossibleActions[minAction];
			//this.printPolicy();
			min = Double.MAX_VALUE;
			minAction = 0;
			
			//Statistics...
			//Add one to number of refreshs...
			this.States[AllStates].addNumRefreshed();
			
		}
		//this.printResidual();
		return this.Policy[CurrentState.id];
	}
	
	Action refreshOnlyCurrentStateV(){
		//Recalculate the V[] Matrix only for current state.

		double min = Double.MAX_VALUE;
		int minAction = 0;		

		for(int i=0; i < this.PossibleActions.length; i++){
			
			double calc = 0;
			
			for(int j=0; j < this.States.length; j++){
				calc += M.Matrix[i][CurrentState.id][j] * this.V[j]; 
			}

			calc += C.Matrix[i][CurrentState.id];

			if(calc < min){
				min = calc;
				minAction = i;
			}
		}
		
		//Calculate residual...
		this.Residual[CurrentState.id] = Math.abs(this.V[CurrentState.id] - min);
		
		this.V[CurrentState.id] = min;
		this.Policy[CurrentState.id] = PossibleActions[minAction];
		
		//Statistics...
		this.CurrentState.addNumRefreshed();
		
		return this.Policy[CurrentState.id];
	}

	void refreshSpecifiedStateV(State s){
		//Recalculate the V[] Matrix only for current state.

		double min = Double.MAX_VALUE;
		int minAction = 0;		

		for(int i=0; i < this.PossibleActions.length; i++){
			
			double calc = 0;
			
			for(int j=0; j < this.States.length; j++){
				calc += M.Matrix[i][s.id][j] * this.V[this.States[j].id]; 
			}
			
			calc = calc * this.gamma; 
			calc += C.Matrix[i][s.id];

			if(calc < min){
				min = calc;
				minAction = i;
			}
		}
		
		//Statistics..
		s.addNumRefreshed();
		
		//Calculate residual...
		this.Residual[s.id] = Math.abs(this.V[s.id] - min);
		
		this.V[s.id] = min;
		this.Policy[s.id] = PossibleActions[minAction];
		
	}
		
	
	
	void printV(){
		System.out.print("V[] = [ ");
		
		for(int i=0; i < this.V.length; i++)
			System.out.print(V[i] + "  ");
		
		System.out.println("]");
	}
	void printStatesRefreshedCount(){
		//this prints how many times each state was refreshed.
		System.out.println("Number of times each state was refreshed:");
		
		for(int i=0; i<this.States.length; i++){
			System.out.print(States[i].numRefreshed + " ");
		}
	}
	
	void printTakenActionsList(){
		System.out.println("Taken Actions List:");
		Iterator<Action> it = this.TakenActions.iterator();
		while(it.hasNext()){
			Action a = it.next();
			System.out.print(a.name + " ");
		}
		
		System.out.println();	
	}
	
	void printPolicy(){
		System.out.println("Policy: ");
		for (int i=0; i< this.Policy.length; i++){
			if(this.Policy[i] != null)	
				System.out.print(this.Policy[i].name + " ");
			else
				System.out.print("null"+ " ");
		}
		System.out.println();
	}
	
	
	void printResidual(){
		System.out.println("Residual: ");
		for (int i=0; i< this.Residual.length; i++){	
				System.out.print(this.Residual[i] + " ");
		}
		System.out.println();
	}
	
	void printLine(){
		for(int i=0; i<80; i++){
			System.out.print("-");
		}
		System.out.println();
	}
	
	void doAction(Action a){
		//Generate random number between 0 and 1.
		Random r = new Random();
		double rand = r.nextDouble();
		
		//Refresh current state
		//System.out.println("Rand Number: "+rand);
		numActionsTaken++;
		
		for(int i=0; ; i++){
			if(rand <= M.Cumulative[a.id][CurrentState.id][i]){
				//System.out.println("Next State: " + i);
				CurrentState = States[i];
				break;
			}
		}
		
		
	}
	
	void printMDP(){
		this.printLine();
		this.M.printMatrix();
		this.printLine();
		this.C.print();
		this.printLine();
		System.out.println("\nInitial State: " + this.CurrentState.id);
		
	}
	
	public void valueIteration(){
		
		this.initializeV();
		this.printV();
		
		boolean AllStatesSolved = false;
		
		
		while(AllStatesSolved == false){
		
			Action nextAction = this.refreshV();
			//this.printV();
			//System.out.println("Taken Action: " + nextAction.name);
			this.doAction(nextAction);
			this.TakenActions.add(nextAction);
			
			//Verifies if all states are solved...
			for(int i=0; i< this.Residual.length; i++){
				if( this.Residual[i] > this.EPSILON){
					AllStatesSolved = false;
					break;
				}
				else{
					AllStatesSolved = true;
				}
			}
			
	
		}
		
		//System.out.println("Meta State achieved in " + this.numActionsTaken + " steps!");
		//this.printTakenActionsList();
		
		this.printPolicy();

	}
	
	public void RTDP(){
		
		this.initializeV();
		this.printV();
		
		while(this.CurrentState.isMeta == false){
		
			Action nextAction = this.refreshOnlyCurrentStateV();
			//this.printV();
			//System.out.println("Taken Action: " + nextAction.name);
			this.doAction(nextAction);
			this.TakenActions.add(nextAction);
		}
		
		//System.out.println("Meta State achieved in " + this.numActionsTaken + " steps!");
		//this.printTakenActionsList();
		this.printPolicy();
	}
	
	private boolean CheckSolved(State s, double epsilon){
		boolean result = true;
		Queue<State> open = new LinkedList<State>();
		Queue<State> closed = new LinkedList<State>();
		
		if(s.solved == false){
			open.add(s);
		}
		
		while(open.isEmpty() == false){
			s = open.poll();
			closed.add(s);
			
			//check residual
			if(this.Residual[s.id] > epsilon){
				result = false;
				continue;
			}
			
			//expand state
			Action a = this.Policy[s.id];
			for(int state_it=0; state_it < this.States.length; state_it++){
				if(this.M.Matrix[a.id][s.id][state_it] > 0){
					if(this.States[state_it].solved == false && 
							open.contains(States[state_it]) == false &&
							closed.contains(States[state_it]) == false){
					
						open.add(States[state_it]);
					}
				}
			}	
		}
		
		if(result == true){
			//label relevant states
			Iterator<State> it = closed.iterator();
			while(it.hasNext()){
				it.next().solved = true;
				
			}
		}
		
		else{
			//update states with residuals and ancestors
			Iterator<State> it = closed.iterator();
			while(it.hasNext()){
				this.refreshSpecifiedStateV(it.next());
				
			}
		}
		
		return result;		
	}
	
	public void initializeLRTDP(){
		for(int i=0; i<this.States.length; i++){
			if(States[i].isMeta == true){
				States[i].solved = true;
			}
		}
	}
	
	public void LRTDP(){
		
		this.initializeV();
		this.initializeLRTDP();
		
		
		//Calculate V() once to avoid null pointer exception...
		
		for(int AllStates=0; AllStates < this.States.length; AllStates++){
			this.refreshSpecifiedStateV(this.States[AllStates]);
		}
		
		
		while(this.InitialState.solved == false){
			this.LRTDPTrial(this.InitialState);
		}

		
		this.printPolicy();
		
	}

	public void LRTDPTrial(State state){
		Queue<State> visited = new LinkedList<State>();
		State s = state;
		//double n = 0;
			
			while(s.solved == false){
				//insert into visited
				visited.add(s);
			
				//check termination at goal states
				if(s.isMeta == true){
					break;
				}
			
				//pick best action and update hash.
				this.refreshSpecifiedStateV(s);
				Action a = this.Policy[s.id];
				this.doAction(a);
				this.TakenActions.add(a);
				s = this.CurrentState;
			
			}
		
			//try labeling visited states in reverse order.

			while(visited.isEmpty() == false){
				s = visited.poll();
				if(this.CheckSolved(s, EPSILON)){
					break;
				}
			}
	
	}
				
	
	public static MDP createRandomMDP(int numEstados, int numAcoes){
		Random r = new Random();
		//Create States...
		//System.out.println("Creating states...");
		
		State Estados[] = new State[numEstados];
		for(int i=0; i < numEstados; i++){
			if(r.nextDouble() > 0.9 && i != 0 || i == numEstados-1){
				//Ensures that at least 1 State is final
				Estados[i] = new State("Estado " + i, i, true);
			}
			else{
				Estados[i] = new State("Estado " + i, i, false);
			}
		}
		
		//Create Actions...
		//System.out.println("Creating Actions...");
		
		Action Acoes[] = new Action[numAcoes];
		for(int i=0; i < numAcoes; i++){
			Acoes[i] = new Action("acao"+i, i);
		}
		
		//Create Transition Matrix...
		//System.out.println("Creating Transition Matrix...");
		
		double a[][][] = new double[numAcoes][numEstados][numEstados];
		
		for(int action=0; action < numAcoes; action++){
			for(int prev_state=0; prev_state < numEstados; prev_state++){
				double cumulative = 0;
				for(int next_state=0; next_state < numEstados; next_state++){
					
					if(cumulative < 1 && next_state < numEstados-1){
						double rand = r.nextDouble();
						/*
						while(rand + cumulative > 1){
							//recalculate random if cumulative > 1.
							rand = r.nextDouble();
						}
						*/
						a[action][prev_state][next_state] = rand*(1-cumulative);
						cumulative = cumulative + a[action][prev_state][next_state];
					}
					else if(cumulative < 1 && next_state == numEstados-1){
						//Last position of next state... Must compute 1 - cumulative
						a[action][prev_state][next_state] = 1 - cumulative;
						cumulative = 1;
					}
					else if(cumulative == 1){
						a[action][prev_state][next_state] = 0;
					}

				}
			
			}
		}
		
		TransitionMatrix MatrizTransicao = new TransitionMatrix(a, numAcoes, numEstados, numEstados);
		
		//Create CostAndReward Matrix...
		
		double c[][] = new double[numAcoes][numEstados];
		
		for(int i=0; i<numAcoes; i++){
			for(int j=0; j<numEstados; j++){
				c[i][j] = r.nextDouble() * 10;
			}
		}
		
		
		CostAndReward Custo = new CostAndReward(c, numAcoes, numEstados);
		
		//Create MDP...
		MDP result = new MDP(Estados, Estados[0], Acoes, MatrizTransicao, Custo);
		return result;
	}

	public void restartMDP(){
		//Method to restart a MDP so a instance of a MDP can be resolved in different algorithms.
		
		//Clean up TakenActions list...
		while(this.TakenActions.isEmpty() == false){
			this.TakenActions.remove();
		}
		
		//Clean up Policy...
		for(int i=0; i < this.Policy.length; i++){
			this.Policy[i] = null;
		}
		//Clean up Residual...
		for(int i=0; i < this.Residual.length; i++){
			this.Residual[i] = 1;
		}
		//Clean up numRefreshed...
		for(int i=0; i < this.States.length; i++){
			this.States[i].numRefreshed = 0;
		}
		
		//Re-initialize V...
		this.initializeV();
		
		//Set current state to initial state...
		this.CurrentState = this.InitialState; 
		
		this.numActionsTaken = 0;
		
	}
	
	public static MDP createSimpleExample(){
		//Create States...
		State Estados[] = new State[2];
		Estados[0] = new State("Estado 0", 0, false);
		Estados[1] = new State("Estado 1", 1, true);
		
		//Create Actions...
		Action Acoes[] = new Action[2];
		Acoes[0] = new Action("acao0", 0);
		Acoes[1] = new Action("acao1", 1);
		
		//Create Transition Matrix...
		double a[][][] = {{{1, 0},{0, 1}}, {{0.4, 0.6},{0.6, 0.4}}};
		TransitionMatrix MatrizTransicao = new TransitionMatrix(a, 2, 2, 2);
		
		//Create CostAndReward Matrix...
		double c[][] = {{1, 0.5}, {1, 2}};
		CostAndReward Custo = new CostAndReward(c, 2, 2);
		
		//Create MDP...
		MDP exemplo = new MDP(Estados, Estados[0], Acoes, MatrizTransicao, Custo);
		return exemplo;
	}
	
	public static MDP createRobotExample(double p){
		//Cannot run RTDP in this example.. 
		
		//Create States...
		State Estados[] = new State[4];
		Estados[0] = new State("S0Y", 0, false);
		Estados[1] = new State("S1Y", 1, false);
		Estados[2] = new State("S0N", 2, false);
		Estados[3] = new State("S1N", 3, true);	
		
		//Create Actions...
		Action Acoes[] = new Action[3];
		Acoes[0] = new Action("goNextRoom", 0);
		Acoes[1] = new Action("getCoffeeAndDeliver", 1);
		Acoes[2] = new Action("stay", 2);
		
		//Create Transition Matrix...
		double a[][][] = { { {0, 1, 0, 0}, {1, 0, 0, 0}, {0, p, 0, 1-p},{p, 0, 1-p, 0} },
						   { {0, 0, 1, 0}, {0, 0, 1, 0}, {0, 0, 1, 0}, {0, 0, 1, 0} },
						   { {1, 0, 0, 0}, {0, 1, 0, 0}, {p, 0, 1-p, 0}, {0,p,0, 1-p}}
						   };
		
		TransitionMatrix MatrizTransicao = new TransitionMatrix(a, 3, 4, 4);
		
		//Create CostAndReward Matrix...
		double c[][] = {{1,1,1,1}, {2,1,4,3},{1,1,0.5,0.5}};
		CostAndReward Custo = new CostAndReward(c, 3, 4);
		
		//Create MDP...
		MDP exemplo = new MDP(Estados, Estados[0], Acoes, MatrizTransicao, Custo);
		return exemplo;		
		
		
		
	}
	
	public static MDP createStoreExample(double p){
		//Cannot run RTDP in this example.. 
		
		//Create States...
		State Estados[] = new State[6];
		Estados[0] = new State("E0V0", 0, false);
		Estados[1] = new State("E1V1", 1, false);
		Estados[2] = new State("E1V0", 2, false);
		Estados[3] = new State("E2V2", 3, true);	
		Estados[4] = new State("E2V1", 4, false);
		Estados[5] = new State("E2V0", 5, false);
		
		//Create Actions...
		Action Acoes[] = new Action[3];
		Acoes[0] = new Action("A0", 0);
		Acoes[1] = new Action("A1", 1);
		Acoes[2] = new Action("A2", 2);
		
		//Create Transition Matrix...
		double a[][][] = { { //Action 0...
							 {1, 0, 0, 0, 0, 0}, 
							 {0, 0.5, 0.5, 0, 0, 0},
							 {0, 0.5, 0.5, 0, 0, 0}, 
							 {0, 0, 0, 0.33, 0.33, 0.34},
							 {0, 0, 0, 0.33, 0.33, 0.34},
							 {0, 0, 0, 0.33, 0.33, 0.34},
							
							},
							
						   //Action 1...
						   { {0, 0.5, 0.5, 0, 0, 0}, 
							 {0, 0, 0, 0.33, 0.33, 0.34}, 
							 {0, 0, 0, 0.33, 0.33, 0.34}, 
							 {0,0,0,0,0,1},
							 {0,0,0,0,0,1},
							 {0,0,0,0,0,1} 
						   },
						   
						   //Action 2...
						   { {0, 0, 0, 0.33, 0.33, 0.34}, 
							 {0,0,0,0,0,1}, 
							 {0,0,0,0,0,1}, 
							 {0,0,0,0,0,1},
							 {0,0,0,0,0,1},
							 {0,0,0,0,0,1} 
						    },						   
						   
						 };
		
		TransitionMatrix MatrizTransicao = new TransitionMatrix(a, 3, 6, 6);
		
		//Create CostAndReward Matrix...
		double c[][] = {{0.5, 0.5, 1, -0.5, 1, 1.5},
						{1, 1, 1.5, 0.5, 1.5, 2},
						{2, 2, 2.5, 1.5, 2.5, 3}
					   };
		CostAndReward Custo = new CostAndReward(c, 3, 6);
		
		//Create MDP...
		MDP exemplo = new MDP(Estados, Estados[0], Acoes, MatrizTransicao, Custo);
		return exemplo;		
		
		
		
	}
	
	public static MDP createMailDeliveryExample(double p, double q, double r){
		//Cannot run RTDP in this example.. 
		
		//Create States...
		State Estados[] = new State[8];
		Estados[0] = new State("NNN", 0, false);
		Estados[1] = new State("YNN", 1, false);
		Estados[2] = new State("NYN", 2, false);
		Estados[3] = new State("NNY", 3, false);	
		Estados[4] = new State("YYN", 4, false);
		Estados[5] = new State("YNY", 5, false);
		Estados[6] = new State("NYY", 6, false);
		Estados[7] = new State("YYY", 7, true);
		
		//Create Actions...
		Action Acoes[] = new Action[3];
		Acoes[0] = new Action("d0", 0);
		Acoes[1] = new Action("d1", 1);
		Acoes[2] = new Action("d2", 2);
		
		//Create Transition Matrix...
		double a[][][] = { { //Action d0...
							 {1-p, p, 0, 0, 0, 0, 0, 0}, 
							 {0, 1, 0, 0, 0, 0, 0, 0},
							 {0, 0, 1-p, 0, p, 0, 0, 0}, 
							 {0, 0, 0, 1-p, 0, p, 0, 0},
							 {0, 0, 0, 0, 1, 0, 0, 0},
							 {0, 0, 0, 0, 0, 1, 0, 0},
							 {0, 0, 0, 0, 0, 0, 1-p, p},
							 {0, 0, 0, 0, 0, 0, 0, 1}
							
							},
							
						   //Action d1...
						   { 
							{1-q, 0, q, 0, 0, 0, 0, 0}, 
							{0, 1-q, 0, 0, q, 0, 0, 0},
							{0, 0, 1, 0, 0, 0, 0, 0}, 
							{0, 0, 0, 1-q, 0, 0, q, 0},
							{0, 0, 0, 0, 1, 0, 0, 0},
							{0, 0, 0, 0, 0, 1-q, 0, q},
							{0, 0, 0, 0, 0, 0, 1, 0},
							{0, 0, 0, 0, 0, 0, 0, 1}
						   },
						   
						   //Action d2...
						   { 
							{1-r, 0, 0, r, 0, 0, 0, 0}, 
							{0, 1-r, 0, 0, 0, r, 0, 0},
							{0, 0, 1-r, 0, 0, 0, r, 0}, 
							{0, 0, 0, 1, 0, 0, 0, 0},
							{0, 0, 0, 0, 1-r, 0, 0, r},
							{0, 0, 0, 0, 0, 1, 0, 0},
							{0, 0, 0, 0, 0, 0, 1, 0},
							{0, 0, 0, 0, 0, 0, 0, 1}   
						   },						   
						   
						 };
		
		TransitionMatrix MatrizTransicao = new TransitionMatrix(a, 3, 8, 8);
		
		//Create CostAndReward Matrix...
		double c[][] = {{10, 9, 8, 6, 7, 5, 4, 3},
						{10, 9, 8, 6, 7, 5, 4, 3},
						{10, 9, 8, 6, 7, 5, 4, 3},	
					   };
		
		CostAndReward Custo = new CostAndReward(c, 3, 8);
		
		//Create MDP...
		MDP exemplo = new MDP(Estados, Estados[0], Acoes, MatrizTransicao, Custo);
		return exemplo;		
		
		
		
	}
	
	public void testAlgorithms(){
		this.printLine();
		System.out.println("LRTDP com residual....");
		this.printLine();
		this.LRTDP();
		this.printV();
		this.printStatesRefreshedCount();
		System.out.println("\n");
		
		this.printLine();
		System.out.println("RTDP....");
		this.printLine();
		this.restartMDP();
		this.RTDP();
		this.printV();
		this.printStatesRefreshedCount();
		System.out.println("\n");
		
		
		this.printLine();
		System.out.println("Value Iteration...");
		this.printLine();
		this.restartMDP();
		this.valueIteration();
		this.printV();
		this.printStatesRefreshedCount();
		System.out.println();
	
	}
	
	public static void main(String args[]){
		
		if(args.length == 0){
			System.out.println("Para executar o problema do robo entregador de cartas:");
			System.out.println("java MDP robo");
			System.out.println("\nPara criar um MDP aleatorio, faca:");
			System.out.println("java MDP random [num. estados] [num. acoes]");
			return;
		}
	
		
		if(args.length == 1){
			
			if(args[0].compareTo("robo") == 0){
				System.out.println("Criando o problema do robo entregador de cartas.");
				MDP exemplo = createMailDeliveryExample(0.9, 0.7, 0.5);
				exemplo.printMDP();
				exemplo.testAlgorithms();
				
			}
			
			else if(args[0].compareTo("random") == 0){
				MDP exemplo = createRandomMDP(15,4);
				System.out.println("Testando para MDP aleatorio com 15 estados e 4 acoes\n");
				exemplo.testAlgorithms();
				
			}			
			
		}
		
		if(args.length == 3){
			if(args[0].compareTo("random") == 0){
				MDP exemplo = createRandomMDP(Integer.parseInt(args[1]),Integer.parseInt(args[2]));
				System.out.println("Testando para MDP aleatorio com " + args[1] + " estados e " + args[2] + " acoes\n");
				exemplo.testAlgorithms();
				
			}			
			
		}
			
	}
}
